# PYTHON_ARGCOMPLETE_OK

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" EeSSH -- SSH client for Eemyle """

import logging
import re
import sys
import traceback

import argcomplete

# Import shell to be sure its parameters will be present in help message and arguments parser
# pylint: disable=unused-import
import eessh.shell  # noqa: F401
from eessh.api import api
from eessh.cache import xdgcache
from eessh.client import EeSSHClient
from eessh.config import config
from eessh.utils import (
    auto_paged,
    generate_server_ansible_inventory_entry,
    generate_server_raw_ssh_command,
    generate_server_ssh_config,
    pretty_format_server_data,
    pretty_format_server_list,
    pretty_format_server_passwords,
    pretty_format_server_tickets,
)

log = logging.getLogger(__name__)


def main(argv=None):
    """EeSSH entry point"""
    parser = config.get_arguments_parser()

    parser.add_argument(
        "-a",
        "--action",
        default="shell",
        choices=["shell", "eeshell", "show", "show-tickets", "show-passwords", "show-comments"],
        help="Action to launch (default: shell)",
    )

    parser.add_argument("-l", "--list", metavar="customer_uid", help="List customer's servers")

    parser.add_argument("-n", "--no-firsthop", action="store_true", help="Skip first hop")

    parser.add_argument(
        "-b", "--broadcast", action="store_true", help="Enable broadcast mode (default is disabled)"
    )

    parser.add_argument(
        "-r", "--resolvable", action="store_true", help="Argument is a resolvable hostname"
    )

    parser.add_argument("-p", "--port", type=int, help="Port for resovable hostname")

    parser.add_argument(
        "-u", "--username", help="Specifies the user to log in as on the remote machine."
    )

    parser.add_argument(
        "-R", "--refresh", action="store_true", help="Force refresh server/customer data"
    )

    parser.add_argument(
        "-I",
        "--ansible-inventory",
        action="store_true",
        help="Generate ansible inventory entries for a customer",
    )

    parser.add_argument(
        "-S",
        "--ssh-command",
        action="store_true",
        help="Generate raw SSH command to connect on a customer host",
    )

    parser.add_argument(
        "-s",
        "--ssh-config",
        action="store_true",
        help="Generate SSH config block to connect on a customer host",
    )

    hostname_arg = parser.add_argument(
        "hostname",
        nargs="?",
        metavar="eeid-hostname",
        help="The hostname (format: [customer UID]-[hostname])",
    )

    def _autocomplete_hostname(prefix, parsed_args, **kwargs):  # pylint: disable=unused-argument
        try:
            config.load_autocompleter_parsed_args(parsed_args)

            prefix_parts = prefix.split("-")
            if len(prefix_parts) <= 1:
                # Autocomplete customer uids
                customer_uid_prefix = prefix if "@" not in prefix else prefix.split("@")[1]
                if parsed_args.refresh or not xdgcache.get("eessh.api:get_customers_uid|"):
                    api.get_customers_uid(refresh=parsed_args.refresh)
                result = [
                    customer_uid if parsed_args.list else f"{customer_uid}-"
                    for customer_uid in api.get_customers_uid()
                    if customer_uid.startswith(customer_uid_prefix)
                ]
                # If only one matching customer_uid, add another fake one to avoid space adding
                # (except in list mode). Note: append_space option normally permit to control it,
                # but it seem buggy
                if len(result) == 1 and not parsed_args.list:
                    result.append(result[0] + " ")
                return result

            # Autocomplete customer server names
            customer_uid = prefix_parts[0]
            servername_prefix = "-".join(prefix_parts[1:])
            if parsed_args.refresh or not xdgcache.get(
                f"eessh.api:get_customer_servernames|{customer_uid}"
            ):
                api.get_customer_servernames(customer_uid, refresh=parsed_args.refresh)
            return (
                f"{customer_uid}-{servername}"
                for servername in api.get_customer_servernames(customer_uid)
                if servername.startswith(servername_prefix)
            )
        except Exception:  # pylint: disable=broad-except
            argcomplete.warn(
                "An exception occurred completing server name:\n" f"{traceback.format_exc()}"
            )
        return None

    hostname_arg.completer = _autocomplete_hostname

    # Parse argument
    options = config.parse_arguments_options(argv)

    # Handle list parameter
    if options.list:
        customer_uid = options.list
        # Get servers list
        servers = api.get_servers(customer_uid, refresh=options.refresh)
        if servers is None:
            parser.error(f"Failed to get servers list for customer {customer_uid}")

        auto_paged(pretty_format_server_list(servers))
        return 0

    # Verify hostname is provided
    if not options.hostname and not options.ansible_inventory and not options.ssh_config:
        parser.error("You must provide a hostname as first positional argument.")

    hostname = options.hostname

    # Handle optional eessh:// prefix to allow handling XDG URI
    if hostname is not None and hostname.startswith("eessh://"):
        hostname = hostname[8:]

    if options.resolvable:
        if "@" in hostname:
            username, hostname = hostname.split("@")
        else:
            username = options.username
    elif hostname is not None:
        matches = re.match(
            r"^((?P<username>[^@]+)@)?(?P<customer_uid>[a-z0-9]{2,3})\-(?P<hostname>.*)$",
            hostname,
        )
        if matches:
            customer_uid = matches.group("customer_uid")
            hostname = matches.group("hostname")
            if matches.group("username") and not options.username:
                options.username = matches.group("username")
        elif options.list or options.ansible_inventory or options.ssh_config:
            customer_uid = hostname
            hostname = None
        else:
            parser.error("Failed to extract customer_uid from hostname")
    else:
        customer_uid = None

    if not api.ready:
        parser.error("Failed to initialize API connector")

    # Raw ssh command generation
    if options.ssh_command:
        ssh_command = generate_server_raw_ssh_command(
            customer_uid, hostname, refresh=options.refresh
        )
        if ssh_command:
            print(ssh_command)
            return 0
        parser.error(
            f"Failed to generate raw SSH command for {hostname} ({customer_uid}). "
            "See log for details."
        )

    # Ansible inventory generation
    if options.ansible_inventory:
        if customer_uid and hostname:
            ansible_entry = generate_server_ansible_inventory_entry(
                customer_uid,
                hostname,
                no_firsthop=options.no_firsthop,
                refresh=options.refresh,
            )
            if ansible_entry:
                print(ansible_entry)
            else:
                parser.error(
                    f"Failed to generate ansible inventory entry for {hostname} ({customer_uid}). "
                    "See log for details."
                )
        elif customer_uid:
            servers = api.get_servers(customer_uid)
            if servers is None:
                parser.error(f"Failed to get servers list for customer {customer_uid}")

            for server in servers:
                if re.match(r"window", server["os_name"], flags=re.IGNORECASE):
                    continue
                ansible_entry = generate_server_ansible_inventory_entry(
                    customer_uid,
                    server["hostname"],
                    no_firsthop=options.no_firsthop,
                )
                if ansible_entry:
                    print(ansible_entry)
        else:
            for customer_uid in api.get_customers_uid():
                servers = api.get_servers(customer_uid)
                if servers is None:
                    parser.error(f"Failed to get servers list for customer {customer_uid}")

                for server in servers:
                    if re.match(r"window", server["os_name"], flags=re.IGNORECASE):
                        continue
                    ansible_entry = generate_server_ansible_inventory_entry(
                        customer_uid,
                        server["hostname"],
                        no_firsthop=options.no_firsthop,
                    )
                    if ansible_entry:
                        print(ansible_entry)

        return 0

    # SSH config generation
    if options.ssh_config:
        if customer_uid and hostname:
            ssh_config = generate_server_ssh_config(
                customer_uid,
                hostname,
                no_firsthop=options.no_firsthop,
                refresh=options.refresh,
            )
            if ssh_config:
                print(ssh_config)
            else:
                parser.error(
                    f"Failed to generate ssh config block for {hostname} ({customer_uid}). "
                    "See log for details."
                )
        elif customer_uid:
            servers = api.get_customer_servernames(customer_uid, refresh=options.refresh)
            if servers is None:
                parser.error(f"Failed to get servers list for customer {customer_uid}")

            for server in servers:
                ssh_config = generate_server_ssh_config(
                    customer_uid,
                    server,
                    no_firsthop=options.no_firsthop,
                )
                if ssh_config:
                    print(ssh_config)
        else:
            for customer_uid in api.get_customers_uid():
                servers = api.get_customer_servernames(customer_uid, refresh=options.refresh)

                for server in servers:
                    ssh_config = generate_server_ssh_config(
                        customer_uid,
                        server,
                        no_firsthop=options.no_firsthop,
                    )
                    if ssh_config:
                        print(ssh_config)

        return 0

    # Use cases where we don't need to launch a client
    if options.action in ("show", "show-tickets", "show-passwords", "show-comments"):
        if options.action == "show-tickets":
            tickets = api.get_tickets(customer_uid, hostname, refresh=options.refresh)
            auto_paged(pretty_format_server_tickets(tickets))
        else:
            data = api.get_data(customer_uid, hostname, refresh=options.refresh)
            if not data:
                print(
                    "[\x1b[31mFAILED\x1b[39m] Fail to retrieve information about server"
                    f" {hostname} ({customer_uid})"
                )
                return 1
            if options.action == "show":
                auto_paged(pretty_format_server_data(data))
            elif options.action == "show-passwords":
                auto_paged(pretty_format_server_passwords(data))
            elif options.action == "show-comments":
                auto_paged(data.get("comments"))
        return 0

    # Flush screen
    sys.stdout.write("\x1b[2J\x1b[H\n\n")
    sys.stdout.flush()

    if options.resolvable:
        data = {
            "customer_uid": None,
            "hostname": hostname,
            "hops": [
                {
                    "host": hostname,
                    "port": options.port or 22,
                    "hostname": hostname,
                    "username": username,
                }
            ],
            "ipmi": None,
        }
    else:
        # Get HOPS
        sys.stdout.write(
            f"[ .... ] Getting connection parameters for {hostname} ({customer_uid}) from GEemyle "
            "server..."
        )
        sys.stdout.flush()

        data = api.get_data(customer_uid, hostname, refresh=options.refresh)
        if not data or not data["hops"]:
            sys.stderr.write(
                "\r[\x1b[31mFAILED\x1b[39m]\n"
                "         ► Are you sure this server exists?\n"
                "         🛈 Run the following command to list existing servers of this customer:\n"
                f"           {sys.argv[0]} -l {customer_uid}"
                "\n\r"
            )
            return 1

        sys.stdout.write("\r[  \x1b[32mOK\x1b[39m  ]\n\r")
        sys.stdout.flush()

    # Launch client
    client = EeSSHClient(
        data,
        agent_forwarding=config.get("connection", "agent_forwarding"),
        broadcast=options.broadcast,
        compress=config.get("connection", "compress"),
        no_firsthop=options.no_firsthop,
        username=options.username,
        x11_forwarding=config.get("connection", "x11_forwarding"),
    )

    client.run(options.action)
