# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Configuration & options parser """

import argparse
import base64
import logging
import os
import stat
import sys
from configparser import ConfigParser
from getpass import getpass
from logging.config import fileConfig
from logging.handlers import SysLogHandler

import argcomplete
import keyring
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from keyring.backend import KeyringBackend
from keyring.backends.fail import Keyring as FailKeyring

from eessh.const import APPNAME, DEFAULT_CONFIG_DIRPATH, DEFAULT_CONFIG_FILENAME, ENCODING, VERSION

try:
    from systemd.journal import JournalHandler
except ImportError:
    JournalHandler = False

log = logging.getLogger(__name__)


class BaseOption:
    """Base configuration option class"""

    def __init__(
        self,
        config,  # pylint: disable=redefined-outer-name
        section,
        name,
        default=None,
        comment=None,
        arg=None,
        short_arg=None,
        arg_help=None,
        **kwargs,
    ):
        self.config = config
        self.section = section
        self.name = name
        self.default = default
        self.comment = comment
        self.arg = arg
        self.short_arg = short_arg
        self.arg_help = arg_help if arg_help else comment
        for kwarg, value in kwargs.items():
            assert hasattr(
                self, kwarg
            ), f"Unsupported {kwarg}={value} pass to {type(self)}.__init__()"
            setattr(self, kwarg, value)

    @property
    def _isset_in_options(self):
        """Check if option is defined in registered arguments parser options"""
        return self.config.options and self._from_options != self.default

    @property
    def _from_options(self):
        """Get option from arguments parser options"""
        value = getattr(self.config.options, self.parser_dest) if self.config.options else None
        log.debug("_from_options(%s, %s) = %s", self.section.name, self.name, value)
        return value

    @property
    def _isset_in_config_file(self):
        """Check if option is defined in the loaded configuration file"""
        return self.config.config_parser and self.config.config_parser.has_option(
            self.section.name, self.name
        )

    @property
    def _from_config(self):
        """Get option value from ConfigParser"""
        return self.config.config_parser.get(self.section.name, self.name)

    def isset(self):
        """Check if option is defined in the loaded configuration file"""
        return self._isset_in_config_file or self._isset_in_options

    def get(self):
        """Get option value from options, config or default"""
        if self._isset_in_options:
            return self._from_options
        if self._isset_in_config_file:
            return self._from_config
        return self.default

    def set(self, value):
        """Set option value to config file"""
        assert (
            self.config.config_parser
        ), "Can set option value only if configuration file is configured"
        if value == "":
            value = None

        if value == self.default or value is None:
            # Remove option from config
            self.config.config_parser.remove_option(self.section.name, self.name)
        else:
            # Store option to config
            if not self.config.config_parser.has_section(self.section.name):
                self.config.config_parser.add_section(self.section.name)

            self.config.config_parser.set(self.section.name, self.name, self.to_config(value))

    @property
    def parser_action(self):
        """Get action as accept by argparse.ArgumentParser"""
        return "store"

    @property
    def parser_type(self):
        """Get type as handle by argparse.ArgumentParser"""
        return str

    @property
    def parser_dest(self):
        """Get option name in arguments parser options"""
        return f"{self.section.name}_{self.name}"

    @property
    def parser_help(self):
        """Get option help message in arguments parser options"""
        if self.arg_help and self.default is not None:
            return f"{self.arg_help} (Default: {self.default})"
        if self.arg_help:
            return self.arg_help
        return None

    @property
    def parser_argument_name(self):
        """Get option argument name in parser options"""
        return (
            self.arg if self.arg else f"--{self.section.name}-{self.name}".lower().replace("_", "-")
        )

    def add_option_to_parser(self, section_opts):
        """Add option to arguments parser"""
        args = [self.parser_argument_name]
        if self.short_arg:
            args.append(self.short_arg)
        kwargs = {
            "action": self.parser_action,
            "dest": self.parser_dest,
            "help": self.parser_help,
            "default": self.default,
        }
        if self.parser_type:  # pylint: disable=using-constant-test
            kwargs["type"] = self.parser_type

        log.debug(
            "add_option_to_parser(%s, %s): argument name(s)=%s / kwargs=%s",
            self.section.name,
            self.name,
            ", ".join(args),
            kwargs,
        )
        section_opts.add_argument(*args, **kwargs)

    def to_config(self, value=None):
        """Format value as stored in configuration file"""
        value = value if value is not None else self.get()
        return "" if value is None else str(value)

    def export_to_config(self):
        """Export option to configuration file"""
        lines = []
        if self.comment:
            lines.append("# " + self.comment)
        value = self.to_config()
        default_value = "" if self.default is None else self.to_config(self.default)
        log.debug(
            "export_to_config(%s, %s): value=%s / default=%s",
            self.section.name,
            self.name,
            value,
            default_value,
        )
        if default_value:
            lines.append(f"# Default: {default_value}")
        if value is not None and value != default_value:
            lines.append(f"{self.name} = {value}")
        else:
            lines.append(f"# {self.name} =")
        lines.append("")
        return "\n".join(lines)


class StringOption(BaseOption):
    """String configuration option class"""


class BooleanOption(BaseOption):
    """Boolean configuration option class"""

    no_arg = None
    no_short_arg = None
    no_arg_help = None

    @property
    def _from_config(self):
        """Get option value from ConfigParser"""
        return self.config.config_parser.getboolean(self.section.name, self.name)

    def to_config(self, value=None):
        """Format value as stored in configuration file"""
        return super().to_config(value).lower()

    @property
    def parser_action(self):
        return "store_true"

    @property
    def parser_type(self):
        return None

    @property
    def parser_help(self):
        """Get option help message in arguments parser options"""
        if self.arg_help and self.default is not None:
            return f"{self.arg_help} (Default: {'enabled' if self.default else 'disabled'})"
        if self.arg_help:
            return self.arg_help
        return None

    @property
    def parser_no_argument_name(self):
        """Get option name for No/False argument in parser options"""
        return self.no_arg or self.parser_argument_name.replace("--", "--no-")

    @property
    def parser_no_dest(self):
        """Get option destination for No/False argument in arguments parser options"""
        return f"{self.section.name}_no_{self.name}"

    @property
    def parser_no_arg_help(self):
        """Get option help message for No/False argument in arguments parser options"""
        if self.no_arg_help and self.default is not None:
            return f"{self.no_arg_help} (Default: {'enabled' if self.default else 'disabled'})"
        if self.no_arg_help:
            return self.no_arg_help
        return None

    def add_option_to_parser(self, section_opts):
        """
        Add options to arguments parser

        Note: Two arguments are use for boolean option, one for Yes/True value and another for
        No/False value. In help message, only the argument that inverse the default value is
        display.
        """
        # Add Yes/True option (with default=None, so we do not use the inherited method)
        args = [self.parser_argument_name]
        if self.short_arg:
            args.append(self.short_arg)
        kwargs = {
            "action": self.parser_action,
            "dest": self.parser_dest,
            "help": self.parser_help if not self.default else argparse.SUPPRESS,
            "default": None,
        }

        log.debug(
            "Boolean.add_option_to_parser(%s, %s): Yes/True argument name(s)=%s / kwargs=%s",
            self.section.name,
            self.name,
            ", ".join(args),
            kwargs,
        )
        section_opts.add_argument(*args, **kwargs)

        # Add the other No/False
        args = [self.no_arg or self.parser_argument_name.replace("--", "--no-", 1)]
        if self.no_short_arg:
            args.append(self.no_short_arg)
        kwargs = {
            "action": self.parser_action,
            "dest": self.parser_no_dest,
            "help": self.parser_no_arg_help if self.default else argparse.SUPPRESS,
            "default": None,
        }

        log.debug(
            "Boolean.add_option_to_parser(%s, %s): No/False argument name(s)=%s / kwargs=%s",
            self.section.name,
            self.name,
            ", ".join(args),
            kwargs,
        )
        section_opts.add_argument(*args, **kwargs)

    @property
    def _from_options(self):
        """Get option from arguments parser options"""
        yes_value = getattr(self.config.options, self.parser_dest) if self.config.options else None
        no_value = (
            getattr(self.config.options, self.parser_no_dest) if self.config.options else None
        )
        log.debug(
            "_from_options(%s, %s) = yes=%s / no=%s",
            self.section.name,
            self.name,
            yes_value,
            no_value,
        )
        if yes_value:
            return True
        if no_value:
            return False
        return self.default

    @property
    def _isset_in_options(self):
        """Check if option is defined in registered arguments parser options"""
        if not self.config.options:
            return False
        return getattr(self.config.options, self.parser_dest) or getattr(
            self.config.options, self.parser_no_dest
        )


class FloatOption(BaseOption):
    """Float configuration option class"""

    @property
    def _from_config(self):
        """Get option value from ConfigParser"""
        return self.config.config_parser.getfloat(self.section.name, self.name)

    @property
    def parser_type(self):
        return float


class IntegerOption(BaseOption):
    """Integer configuration option class"""

    @property
    def _from_config(self):
        """Get option value from ConfigParser"""
        return self.config.config_parser.getint(self.section.name, self.name)

    def to_config(self, value=None):
        """Format value as stored in configuration file"""
        value = value if value is not None else self.get()
        return str(int(value)) if value is not None else ""

    @property
    def parser_type(self):
        return int


class PasswordOption(StringOption):
    """Password configuration option class"""

    def __init__(self, *arg, username_option=None, keyring_value=None, **kwargs):
        super().__init__(*arg, **kwargs)
        self.username_option = username_option
        self.keyring_value = keyring_value if keyring_value is not None else "keyring"

    @property
    def _keyring_service_name(self):
        """Return keyring service name"""
        return ".".join([self.config.shortname, self.section.name, self.name])

    @property
    def _keyring_username(self):
        """Return keyring username"""
        return self.section.get(self.username_option) if self.username_option else self.name

    def get(self):
        """Get option value"""
        value = super().get()

        if value != self.keyring_value:
            return value

        service_name = self._keyring_service_name
        username = self._keyring_username
        log.debug("Retrieve password %s for username=%s from keyring", service_name, username)
        value = self.config.keyring.get_password(service_name, username)

        if value is None:
            value = getpass(
                # pylint: disable=consider-using-f-string
                "Please enter %s%s: "
                % (
                    self.comment if self.comment else f"{self.section.name} {self.name}",
                    f" for {username}" if username != self.name else "",
                )
            )
            self.set(value, force_using_keyring=True)
        return value

    def to_config(self, value=None):
        """Format value as stored in configuration file"""
        value = value if value is not None else super().get()
        if value == self.keyring_value:
            return self.keyring_value
        return super().to_config(value)

    def set(self, value, force_using_keyring=False):
        """Set option value to config file"""
        if force_using_keyring or super().get() == self.keyring_value:
            self.config.keyring.set_password(
                self._keyring_service_name, self._keyring_username, value
            )
            value = self.keyring_value
        super().set(value)


class ConfigSection:
    """Configuration section class"""

    # pylint: disable=redefined-outer-name
    def __init__(self, config, name, comment=None, order=None):
        self.config = config
        self.name = name
        self.options = {}
        self.comment = comment
        self.order = order

    def add_option(self, _type, name, **kwargs):
        """
        Add option

        :param _type: Option type, derivated from BaseOption
        :param name: Option name
        :param **kwargs: Dict of raw option for type class
        """
        assert not self.defined(name), f"Duplicated option {name}"
        self.options[name] = _type(self.config, self, name, **kwargs)
        return self.options[name]

    def defined(self, option):
        """Check if option is defined"""
        return option in self.options

    def isset(self, option):
        """Check if option is set"""
        return self.defined(option) and self.options[option].isset()

    def get(self, option):
        """Get option value"""
        assert self.defined(option), f"Option {option} unknown"
        return self.options[option].get()

    def set(self, option, value):
        """Set option value"""
        assert self.defined(option), f"Option {option} unknown"
        return self.options[option].set(value)

    def add_options_to_parser(self, parser):
        """Add section to argparse.ArgumentParser"""
        assert isinstance(parser, argparse.ArgumentParser)
        section_opts = parser.add_argument_group(
            self.comment if self.comment else self.name.capitalize()
        )

        for option in self.options:  # pylint: disable=consider-using-dict-items
            self.options[option].add_option_to_parser(section_opts)

    def export_to_config(self):
        """Export section and their options to configuration file"""
        lines = []
        if self.comment:
            lines.append(f"# {self.comment}")
        lines.append(f"[{self.name}]")
        for option in self.options:  # pylint: disable=consider-using-dict-items
            lines.append(self.options[option].export_to_config())
        return "\n".join(lines)


class Config:
    """Configuration helper"""

    default_filename = DEFAULT_CONFIG_FILENAME
    default_directory_path = DEFAULT_CONFIG_DIRPATH
    version = VERSION
    encoding = ENCODING

    def __init__(self, appname, shortname=None, **kwargs):
        self.appname = appname
        self.shortname = shortname if shortname is not None else __name__.split(".", maxsplit=1)[0]
        for arg, value in kwargs.items():
            assert hasattr(self, arg), f"Config: Invalid argument {arg}"
            setattr(self, arg, value)

        self.config_parser = None
        self.options_parser = None
        self.options = None
        self.sections = {}
        self._loaded_callbacks = []
        self._filepath = None
        self._keyring = None
        self._children = []

    def add_child(self, child):
        """Add child config object"""
        assert isinstance(child, Config)
        self._children.append(child)

    def add_section(self, name, loaded_callback=None, **kwargs):
        """
        Add section

        : param name: The section name
        : param loaded_callback: An optional callback method that will be executed after configuration is loaded
                                The specified callback method will receive Config object as parameter.
        : param ** kwargs: Raw parameters dict pass to ConfigSection __init__() method
        """
        assert name not in self.sections, f"Duplicated section {name}"

        self.sections[name] = ConfigSection(self, name, **kwargs)
        if loaded_callback:
            self._loaded_callbacks.append(loaded_callback)
            # If configuration is already loaded, execute callback immediately
            if self._filepath:
                loaded_callback(self)
        return self.sections[name]

    def remove_section(self, name):
        """Remove existing section"""
        assert name in self.sections, f"Unknown section {name}"
        del self.sections[name]

    def defined(self, section, option):
        """Check option is defined in specified section"""
        return section in self.sections and self.sections[section].defined(option)

    def isset(self, section, option):
        """Check option is set in specified section"""
        return section in self.sections and self.sections[section].isset(option)

    def get(self, section, option):
        """Get option value"""
        assert self.config_parser or self.options, "Unconfigured options parser"
        assert self.defined(section, option), f"Unknown option {section}.{option}"
        value = self.sections[section].get(option)
        log.debug("get(%s, %s): %s (%s)", section, option, value, type(value))
        return value

    def set(self, section, option, value):
        """Set option value"""
        assert self.config_parser, "Unconfigured options parser"
        assert self.defined(section, option), f"Unknown option {section}.{option}"
        self.sections[section].set(option, value)

    def _execute_loaded_callbacks(self):
        """Execute loaded callbacks"""
        for callback in self._loaded_callbacks:
            callback(self)

    def load_file(self, filepath):
        """Read configuration file"""

        self.config_parser = ConfigParser()
        self._filepath = filepath

        # Checking access of configuration file
        if not os.path.isfile(filepath):
            return True

        if not os.access(filepath, os.R_OK):
            return False

        try:
            self.config_parser.read(filepath, encoding=self.encoding)
        except Exception:  # pylint: disable=broad-except
            self.config_parser = None
            log.exception("Failed to read configuration file %s", filepath)
            return False

        # Logging initialization
        if self.config_parser.has_section("loggers"):
            fileConfig(filepath)
        else:
            # Otherwise, use systemd journal (or syslog) handler
            handler = (
                JournalHandler(SYSLOG_IDENTIFIER=self.shortname)
                if JournalHandler
                else SysLogHandler()
            )
            handler.setFormatter(
                logging.Formatter("%(levelname)s | %(name)s:%(lineno)d | %(message)s")
            )
            logging.getLogger().addHandler(handler)

        self._filepath = filepath

        for child in self._children:
            child.load_file(filepath)

        return True

    def save(self, filepath=None):
        """Save configuration file"""
        filepath = filepath if filepath else self._filepath
        assert filepath, "Configuration filepath is not set or provided"

        # Checking access of target directory
        dirpath = os.path.dirname(filepath)
        if not os.path.exists(dirpath):
            log.info('Configuration directory "%s" does not exist, try to create it', dirpath)
            os.mkdir(dirpath)

        if not os.path.isdir(dirpath) or not os.access(dirpath, os.R_OK | os.W_OK | os.X_OK):
            log.error('Configuration directory "%s" does not exist (or not writable)', dirpath)
            return False

        if os.path.isfile(filepath) and not os.access(filepath, os.W_OK):
            log.error('Configuration file "%s" is not writable', filepath)
            return False

        # pylint: disable=consider-using-f-string
        lines = ["#\n# %s configuration\n#\n" % self.appname]

        for section_name in self.ordered_section_names:
            lines.append("")
            lines.append(self.sections[section_name].export_to_config())

        for child in self._children:
            for section_name in child.ordered_section_names:
                lines.append("")
                lines.append(child.sections[section_name].export_to_config())

        try:
            with open(filepath, "wb") as file_desc:
                file_desc.write("\n".join(lines).encode(self.encoding))

            # Privacy!
            os.chmod(filepath, stat.S_IRUSR | stat.S_IWUSR)
        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to write generated configuration file %s", filepath)
            return False
        return True

    @property
    def ordered_section_names(self):
        """Get ordered list of section names"""
        return sorted(self.sections.keys(), key=lambda section: self.sections[section].order)

    def get_arguments_parser(self, **kwargs):
        """Get arguments parser"""
        if self.options_parser:
            return self.options_parser

        self.options_parser = argparse.ArgumentParser(
            description=f"{self.appname} (version: {VERSION})", **kwargs
        )

        self.options_parser.add_argument(
            "-c",
            "--config",
            default=self.config_filepath,
            help=f"Configuration file to use (default: {self.config_filepath})",
        )

        self.options_parser.add_argument(
            "--save",
            action="store_true",
            dest="save",
            help="Save current configuration to file",
        )

        self.options_parser.add_argument(
            "-d", "--debug", action="store_true", help="Show debug messages"
        )

        self.options_parser.add_argument(
            "-v", "--verbose", action="store_true", help="Show verbose messages"
        )

        self.options_parser.add_argument(
            "--console", action="store_true", help="Log messages on console"
        )

        self.options_parser.add_argument("-L", "--log-file", help="Log file path")

        self.add_options_to_parser(self.options_parser)

        for child in self._children:
            child.add_options_to_parser(self.options_parser)

        return self.options_parser

    logfile_handler = None
    console_handler = None

    def initialize_logging(self, options):
        """Initialize logging"""
        if options.debug:
            logging.getLogger().setLevel(logging.DEBUG)
        elif options.verbose:
            logging.getLogger().setLevel(logging.INFO)

        if options.log_file:
            if self.logfile_handler and self.logfile_handler.baseFilename == options.log_file:
                log.debug(
                    "Disable log file handler on %s before replacing it",
                    self.logfile_handler.baseFilename,
                )
                logging.getLogger().removeHandler(self.logfile_handler)
                self.logfile_handler.close()
                self.logfile_handler = None
            if not self.logfile_handler:
                self.logfile_handler = logging.FileHandler(options.log_file)
                self.logfile_handler.setFormatter(
                    logging.Formatter(
                        "%(asctime)s - %(module)s:%(lineno)d - %(levelname)s - %(message)s"
                    )
                )
                logging.getLogger().addHandler(self.logfile_handler)
                log.debug("Enable logging in log file %s", options.log_file)
        elif self.logfile_handler:
            log.debug("Disable log file handler on %s", self.logfile_handler.baseFilename)
            logging.getLogger().removeHandler(self.logfile_handler)
            self.logfile_handler.close()
            self.logfile_handler = None

        if options.console and not self.console_handler:
            self.console_handler = logging.StreamHandler(sys.stderr)
            self.console_handler.setFormatter(
                logging.Formatter("%(name)s:%(lineno)d - %(levelname)s - %(message)s")
            )
            logging.getLogger().addHandler(self.console_handler)
            log.debug("Enable logging on console")
        elif not options.console and self.console_handler:
            log.debug("Disable logging on console")
            logging.getLogger().removeHandler(self.console_handler)
            self.console_handler.close()
            self.console_handler = None

    def parse_arguments_options(self, argv=None, parser=None, create=True, exit_after_created=True):
        """
        Parse arguments options

        :param argv: Optional arguments list to parse (default: sys.argv[1:])
        :param parser: Optional argparse.ArgumentParser use
                       (default: generated by self.get_arguments_parser())
        :param create: If True, configuration file will be created if it does not exits
                       (default: True)
        :param exit_after_created: If True, script will end after configuration file
                        creation (default: True)
        """
        parser = parser if parser else self.get_arguments_parser()
        argcomplete.autocomplete(parser)
        options = parser.parse_args(argv if argv else sys.argv[1:])

        self.initialize_logging(options)

        if options.config:
            options.config = os.path.abspath(options.config)

            already_saved = False
            if not os.path.isfile(options.config) and (create or options.save):
                log.warning("Configuration file is missing, generate it (%s)", options.config)
                # Make sure to load provided options to keep it in saved file
                self.load_options(options)
                self.save(options.config)
                if exit_after_created:
                    sys.exit(0)
                already_saved = True

            # Load configuration file
            if os.path.isfile(options.config) and not self.load_file(options.config):
                parser.error(f"Failed to load configuration from file {options.config}")

            if options.save and not already_saved:
                # Make sure to load provided options to keep it in saved file
                self.load_options(options)
                self.save()
                sys.exit(0)

        self.load_options(options)
        self._execute_loaded_callbacks()
        return options

    def load_options(self, options):
        """Register arguments parser options"""
        assert isinstance(options, argparse.Namespace)
        self.options = options
        log.debug("Argument options: %s", options)
        for child in self._children:
            child.load_options(options)

    def load_autocompleter_parsed_args(self, options):
        """Load parsed arguments provide to autocompleter methods"""
        self.initialize_logging(options)
        if options.config:
            self.load_file(options.config)
        self.load_options(options)
        self._execute_loaded_callbacks()

    def add_options_to_parser(self, parser):
        """Add sections and their options to parser"""
        for section in self.ordered_section_names:
            self.sections[section].add_options_to_parser(parser)

    @property
    def config_dir(self):
        """Retrieve configuration directory path"""
        return os.path.dirname(self._filepath) if self._filepath else self.default_directory_path

    @property
    def config_filepath(self):
        """Retrieve configuration file path"""
        if self._filepath:
            return self._filepath
        return os.path.join(self.config_dir, self.default_filename)

    @property
    def keyring(self):
        """Retrieve password keyring"""
        if self._keyring:
            return self._keyring
        self._keyring = keyring.get_keyring()
        if isinstance(self._keyring, FailKeyring):
            filepath = os.path.join(
                os.path.dirname(self.config_filepath),
                f".{os.path.basename(self.config_filepath)}.keyring",
            )
            log.warning(
                "No supported keyring backend found, use alternative backend and store collected "
                "passwords in file '%s'.",
                filepath,
            )
            self._keyring = ConfigAltKeyringBackend(filepath)
        return self._keyring


class ConfigAltKeyringBackend(KeyringBackend):
    """
    An alternative keyring backend used if no native keyring is found on system.
    """

    priority = 1

    _file = None
    _filepath = None
    _key_section = "key"
    _key_option = "value"
    _key_salt_option = "salt"
    _key_store_option = "ask_to_store_it"
    _encoding = ENCODING

    def __init__(self, filepath):
        super().__init__()
        self._file = ConfigParser()
        self.load_file(filepath)

        if not self._file.has_section(self._key_section):
            self._file.add_section(self._key_section)

        if not self._file.has_option(self._key_section, self._key_option):
            key = self.password_to_key(self.ask_for_password())
            self.store_key(key)
        else:
            key = self._file.get(self._key_section, self._key_option)
            if not key:
                key = self.password_to_key(self.ask_for_password())
                self.store_key(key)

        self._fernet = Fernet(key)

    @staticmethod
    def ask_for_password():
        """Ask user to enter keyring password"""
        return getpass("Please enter EeSSH keyring password:")

    def password_to_key(self, password):
        """Convert user password to Fernet key"""
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=self.get_salt(),
            iterations=480000,
        )
        return base64.urlsafe_b64encode(kdf.derive(password.encode(self._encoding)))

    def get_salt(self):
        """Retrieve or generate and store the salt to derive key from user password"""
        if self._file.has_option(self._key_section, self._key_salt_option):
            return base64.urlsafe_b64decode(
                self._file.get(self._key_section, self._key_salt_option)
            )
        self._file.set(
            self._key_section,
            self._key_salt_option,
            base64.urlsafe_b64encode(os.urandom(16)).decode(),
        )
        self.save()
        return self.get_salt()

    def store_key(self, key):
        """Store key in file"""
        if self._file.has_option(
            self._key_section, self._key_store_option
        ) and not self._file.getboolean(self._key_section, self._key_store_option):
            return
        while True:
            response = input("Do you want to store it for the next time [y/N]? ").strip().lower()
            if response not in ["y", "n", ""]:
                print("Invalid answer")
            else:
                self._file.set(
                    self._key_section, self._key_option, key.decode() if response == "y" else ""
                )
                if response in ["n", ""]:
                    while True:
                        response = input("Don't ask to store it [Y/n]? ").strip().lower()
                        if response not in ["y", "n", ""]:
                            print("Invalid answer")
                        else:
                            self._file.set(
                                self._key_section,
                                self._key_store_option,
                                str(response not in ["y", ""]).lower(),
                            )
                            break
                self.save()
                break

    def load_file(self, filepath=None):
        """Load file"""
        filepath = filepath if filepath else self._filepath
        # Checking access of target file/directory
        dirpath = os.path.dirname(filepath)
        if os.path.isfile(filepath):
            if not os.access(filepath, os.W_OK):
                log.error(
                    "Alternative keyring backend file %s is not writable. Only store collected "
                    "password in memory.",
                    filepath,
                )
            else:
                try:
                    self._file.read(filepath, encoding=self._encoding)
                    self._filepath = filepath
                except Exception:  # pylint: disable=broad-except
                    log.exception(
                        "Failed to load alternative keyring backend file %s. Only store collected "
                        "password in memory.",
                        filepath,
                    )
        elif not os.path.isdir(dirpath) or not os.access(dirpath, os.R_OK | os.W_OK | os.X_OK):
            log.error(
                "Directory %s does not exist (or not writable). Only store collected passwords in "
                "memory.",
                dirpath,
            )
        else:
            self._filepath = filepath

    def save(self):
        """Save changes in backend file"""
        if not self._filepath:
            return
        try:
            with open(self._filepath, "w", encoding=self._encoding) as file_desc:
                self._file.write(file_desc)
            # Privacy!
            os.chmod(self._filepath, stat.S_IRUSR | stat.S_IWUSR)
            log.debug("Keyring backend file %s updated.", self._filepath)
        except Exception:  # pylint: disable=broad-except
            log.exception("Fail to save alternative keyring backend file %s.")

    def set_password(self, service, username, password):
        """Set a password"""
        assert service != self._key_section, f"Service name '{self._key_section}' is reserved."
        if not self._file.has_section(service):
            self._file.add_section(service)

        self._file.set(
            service, username, self._fernet.encrypt(password.encode(self._encoding)).decode()
        )
        self.save()

    def get_password(self, service, username):
        """Get a password"""
        if not self._file.has_section(service) or not self._file.has_option(service, username):
            return None

        try:
            password = self._fernet.decrypt(self._file.get(service, username))
        except InvalidToken:
            print("Invalid password provided for EeSSH keyring.")
            key = self.password_to_key(self.ask_for_password())
            self.store_key(key)
            self._fernet = Fernet(key)
            return self.get_password(service, username)

        return password.decode(self._encoding)

    def delete_password(self, service, username):
        """Remove a password"""
        if self._file.has_option(service, username):
            self._file.remove_option(service, username)
        self.save()


# Initialize config object
config = Config(APPNAME, version=VERSION, encoding=ENCODING)
