# EeSSH -- SSH client for Eemyle
# By: Benjamin Renard <brenard@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Eemyle API 1.0 connector """

import logging
import threading

from requests import JSONDecodeError, RequestException, Session

from eessh.cache import xdgcache as region
from eessh.const import VERSION

log = logging.getLogger(__name__)


class Eemyle10ConnectorError(RequestException):
    """
    Eemyle API 1.0 connector error
    """

    def __init__(self, err):
        super().__init__(self)
        self.err = err

    def __repr__(self):
        return f"<Eemyle10ConnectorError: {self.err}>"


class Eemyle10Connector:
    """
    Eemyle API 1.0 connector
    """

    def __init__(
        self, url, username=None, password=None, debug=False, **kwargs
    ):  # pylint: disable=unused-argument
        self.url = url
        self.__error = None
        self.__session = Session()
        self.__session.auth = (username, password)
        self.__session.headers["Content-type"] = "application/json"

        self.__lock = threading.Lock()

    def __enter__(self):
        """Acquire lock"""
        self.__lock.acquire()
        return self

    def __exit__(self, type_, value, traceback):
        """Release lock"""
        self.__lock.release()

    @property
    def error(self):
        """Retrieve last error"""
        return self.__error

    @region.cache_on_arguments_if_not_refresh()
    def get_customers(self):
        """Retrieve customers info"""
        result = self.query("customers")
        if result is None or "customers" not in result:
            return None

        customers = []
        for eeid, customer in result["customers"].items():
            if not customer["has_server"]:
                continue

            customers.append(
                {
                    "customer_uid": eeid,
                    "name": customer["name"],
                    "city": customer["city"],
                    "address": customer["street"],
                    "phone": customer["phone"],
                    "has_server": customer["has_server"],
                }
            )

        return customers

    @region.cache_on_arguments_if_not_refresh(leave_refresh_kwarg=True)
    def _get_server_id(self, customer_uid, hostname, refresh=False):
        """Retrieve server ID from customer UID and server hostname"""
        # pylint: disable=unexpected-keyword-arg
        servers = self.get_servers(customer_uid, refresh=refresh)
        if servers is None:
            log.warning(
                "_get_server_id(%s, %s): failed to list customer servers", customer_uid, hostname
            )
            return None

        for server in servers:
            if server["hostname"] == hostname:
                return str(server["id"])

        log.warning(
            "_get_server_id(%s, %s): server not found in list of customer servers",
            customer_uid,
            hostname,
        )
        return None

    def get_data(self, customer_uid=None, hostname=None, server_id=None, refresh=False):
        """Retrieve server extended info"""
        if refresh:
            # Invalidate server caches
            self._invalidate_server_cache(
                customer_uid=customer_uid, hostname=hostname, server_id=server_id
            )

        # Get server
        return (
            self.get_server(server_id=server_id)
            if server_id
            else self.get_server(customer_uid=customer_uid, hostname=hostname)
        )

    def _invalidate_server_cache(self, customer_uid=None, hostname=None, server_id=None):
        """Invalidate server data cached"""
        if server_id:
            log.debug("Invalidate server #%s cache", server_id)
            self.get_server.invalidate(self, server_id=server_id)  # pylint: disable=no-member
        if customer_uid and hostname:
            log.debug("Invalidate server %s-%s cache", customer_uid, hostname)
            # pylint: disable=no-member
            self.get_server.invalidate(self, customer_uid=customer_uid, hostname=hostname)

    @region.cache_on_arguments()
    def get_server(self, customer_uid=None, hostname=None, server_id=None):
        """Retrieve server info"""
        if server_id:
            assert server_id and str(server_id) not in (
                "0",
                "-1",
            ), f"Invalid server ID '{server_id}'"
            result = self.query(
                f"servers/{server_id}", details=["interfaces", "users", "connection_hops"]
            )
        else:
            assert (
                customer_uid and hostname
            ), f"Invalid customer UID '{customer_uid}' & hostname '{hostname}'"
            result = self.query(
                f"servers/{customer_uid}/{hostname}",
                details=["interfaces", "users", "connection_hops"],
            )
        if result is None or "server" not in result:
            return None
        log.debug("get_server(%s): API data=%s", server_id, result)

        server = result["server"]
        data = {
            "id": server["id"],
            "customer_uid": server["id_customer"],
            "hostname": server["name"].split()[0],
            "description": server["usage"],
            "comments": server["comments"],
            "security": server["security_check"],
            "interfaces": list(server["interfaces"].values()),
            "connection_interface": server["interfaces"].get(server["id_connection_interface"]),
            "physical_host": (
                self.get_server(server_id=server["id_server_physical_host"])
                if server["id_server_physical_host"] and server["id_server_physical_host"] > 0
                else None
            ),
            "users": list(server["users"].values()),
        }

        # Compute connection hops info
        data["hops"] = []
        for connection_hop in server["connection_hops"] if server["connection_hops"] else []:
            hop = {
                "hostname": connection_hop.get("server_name"),
                "host": connection_hop.get("ip"),
                "port": int(connection_hop.get("ssh_port", "22")),
                "username": connection_hop.get("username"),
                "password": connection_hop.get("password") or None,
            }

            if connection_hop["root_password"] == "sudo" or connection_hop["sudo_root"]:
                hop["su_method"] = "sudo"
                hop["su_username"] = connection_hop["root_username"] or "root"
                hop["su_password"] = connection_hop["password"] or None

            elif connection_hop["root_username"] == "enable":
                hop["su_method"] = "enable"
                hop["su_password"] = connection_hop["root_password"] or None

            else:
                hop["su_method"] = "su"
                hop["su_username"] = connection_hop["root_username"]
                hop["su_password"] = connection_hop["root_password"] or None

            data["hops"].append(hop)

        # Compute IPMI info
        ipmi_interface = server["interfaces"].get("id_ipmi_interface")
        data["ipmi"] = {
            "host": ipmi_interface["ip"] if ipmi_interface else None,
            "port": server.get("ipmi_port", 623),
            "username": server["ipmi_user_name"] or None,
            "password": server["ipmi_user_password"] or None,
        }

        data["ipmi"]["enabled"] = (
            server["ipmi_connectable"]
            and data["ipmi"]["host"]
            and data["ipmi"]["username"] is not None
            and data["ipmi"]["password"] is not None
        )

        return data

    def update_data(self, changes, customer_uid=None, hostname=None, server_id=None):
        """Update server info"""

        mapped_changes = {}
        simple_map = {"hostname": "name", "description": "usage", "comments": "comments"}
        for key, value in changes.items():
            if key not in simple_map:
                log.error(
                    'update_data(%s): change on "%s" not supported/invalid key', server_id, key
                )
                return False
            mapped_changes[simple_map[key]] = value

        log.debug("Changes: %s", mapped_changes)
        if server_id:
            assert server_id and str(server_id) not in (
                "0",
                "-1",
            ), f"Invalid server ID '{server_id}'"
            result = self.query(f"servers/{server_id}/modify", method="POST", **mapped_changes)
        else:
            assert (
                customer_uid and hostname
            ), f"Invalid customer UID '{customer_uid}' & hostname '{hostname}'"
            result = self.query(
                f"servers/{customer_uid}/{hostname}/modify", method="POST", **mapped_changes
            )
        log.debug("update_data(%s): result = %s", server_id, result)
        if not isinstance(result, dict) or not result.get("success"):
            return False

        # Invalidate cache data of server
        self._invalidate_server_cache(
            customer_uid=customer_uid, hostname=hostname, server_id=server_id
        )

        return True

    @region.cache_on_arguments_if_not_refresh()
    def get_servers(self, customer_uid):
        """Retrieve customer servers"""
        if not customer_uid:
            return None

        result = self.query(f"servers/{customer_uid}", details=["users", "usages"])
        if result is None or "servers" not in result:
            return None

        servers = []
        for server_id, server in result["servers"].items():
            if server["dead"]:
                continue

            servers.append(
                {
                    "id": server_id,
                    "hostname": server["name"].split()[0],
                    "hoster": None,
                    "os_name": server["os"],
                    "description": server["usage"],
                    "security": server["security_check"],
                    "usage": ", ".join(
                        sorted(server["usages"].values()) if server["usages"] else []
                    ),
                    "usernames": (
                        sorted([user["username"] for user in server["users"].values()])
                        if server["users"]
                        else []
                    ),
                }
            )

        return servers

    @region.cache_on_arguments_if_not_refresh(leave_refresh_kwarg=True)
    def get_tickets(self, customer_uid=None, hostname=None, server_id=None, refresh=False):
        """Retrieve server tickets"""
        if server_id is None:
            server_id = self._get_server_id(customer_uid, hostname, refresh=refresh)
            if server_id is None:
                return None

        result = self.query(f"servers/{server_id}/tickets")
        if result is None or "tickets" not in result:
            return None

        tickets = []
        for data in result["tickets"]:
            tickets.append(data)

        return tickets

    def query(self, relative_url, method=None, **data):
        """Sends a request to the server"""
        method = method or "GET"
        url = f"{self.url}/api/1.0/{relative_url}"

        log.debug(
            "Make a %s request on URL %s %s",
            method,
            url,
            f"with data = {data}" if data else "without data",
        )

        self.__error = None

        try:
            request = self.__session.request(
                method,
                url,
                headers={"User-Agent": f"EeSSH/{VERSION}"},
                json=data,
            )
            request.raise_for_status()
            result = request.json()
            if not result.get("success"):
                log.error(
                    "Error occurred calling Eemyle API 1.0: method=%s, URL=%s, data=%s, result=%s",
                    method,
                    url,
                    data,
                    result,
                )
                self.__error = Eemyle10ConnectorError(
                    result.get(
                        "error",
                        " / ".join(result["errors"]) if "errors" in result else "Unknown error",
                    )
                )
        except JSONDecodeError as err:
            log.exception(
                "Error occurred decoding Eemyle API 1.0 JSON response: method=%s, URL=%s, data=%s, response: %s",
                method,
                url,
                data,
                result.text if "result" in locals() else "[not set]",
            )
            self.__error = Eemyle10ConnectorError(err)
            result = None
        except RequestException as err:  # pylint: disable=broad-except
            log.exception(
                "Error occurred calling Eemyle API 1.0: method=%s, URL=%s, data=%s",
                method,
                url,
                data,
            )
            self.__error = Eemyle10ConnectorError(err)
            result = None
        except Exception as err:  # pylint: disable=broad-except
            log.exception(
                "Unexpected exception occurred calling Eemyle API 1.0: method=%s, URL=%s, data=%s",
                method,
                url,
                data,
            )

            # Unknown error
            self.__error = Eemyle10ConnectorError(err)
            result = None

        return result
